#ifndef _AUDIOIO_WAVE_H
#define _AUDIOIO_WAVE_H

#include <string>
#include "audioio.h"
#include "samplebuffer.h"

#include "eca-fileio.h"

typedef struct {
    unsigned short int format;
    unsigned short int channels;
    unsigned int long srate;
    unsigned int long byte_second;
    unsigned short int align;
    unsigned short int bits;
} RF;

typedef struct {
    char sig[4];
    unsigned long bsize;
} RB;

typedef struct {
    char id[4];
    unsigned long size;
    char wname[4];
} RH;

class WAVEFILE : public AUDIO_IO_DEVICE {
  //
  // Represent a RIFF WAVE -file (.wav)
  //
  // This class currently supports only a limited set of features:
  //     - format 1 sample data: Pulse Code Modulation (PCM) Format
  //         - only 8 and 16 bit data supported
  //         - if 16 bits, least significant byte first as specified
  //           in the stantard
  //

  ECA_FILE_IO* fio;

  RH riff_header;
  RF riff_format;

  long int data_start_position;
  bool double_buffering_rep;
    
  void position_in_samples_hook(void);

 public:

  void   get_sample(SAMPLE_BUFFER* t);
  void   put_sample(SAMPLE_BUFFER* t);

  void open_device(void);
  void close_device(void);

  inline bool is_realtime(void) const { return(false); }
    
  WAVEFILE (const string& name, const SIMODE mode, const AIO_PARAMS& form, int bsize, bool double_buffering = false);
  WAVEFILE::~WAVEFILE(void);
  WAVEFILE* clone(void) { return new WAVEFILE(*this); }    

 private:

  WAVEFILE(const WAVEFILE& x) { cerr << "WAVE construct"; }
  WAVEFILE& operator=(const WAVEFILE& x) {  return(*this); }    

  void update(void);        
  void set_length_in_bytes(void);
  void read_riff_header (void);
  bool next_riff_block(RB *t, unsigned long int *offtmp);
  void read_riff_fmt(void);
  void write_riff_header (void);
  void write_riff_fmt(void);
  void write_riff_datablock(void);
  void update_riff_datablock(void);
  void find_riff_datablock (void);
  signed long int find_block(const char* fblock);
};

#endif
